import sys
import time
from typing import Tuple, List, Any, Union


from .color import Color
from .parser import ColorParser
from .fore import Fore

__all__ = ["gradient", "gradient_print", "gradient_change", "gradient_scroll"]


def gradient(start: Tuple[int, int, int], end: Tuple[int, int, int], steps: int) -> List[Tuple[int, int, int]]:
    """
    generate gradients with start color, end color and amount of steps.
    -------------------------------------------------------------------
    :param start: start color, this must be a tuple of 3 ints.
    :param end: end color, this must be a tuple of 3 ints.
    :param steps: amount of steps.
    """
    rs = [start[0]]
    gs = [start[1]]
    bs = [start[2]]
    for step in range(1, steps):
        rs.append(round(start[0] + (end[0]-start[0]) * step/steps))
        gs.append(round(start[1] + (end[1]-start[1]) * step/steps))
        bs.append(round(start[2] + (end[2]-start[2]) * step/steps))

    return list(zip(rs, gs, bs))


def gradient_print(
        *values: Any,
        start_color: Union[str, int, Tuple[int, int, int], Color],
        end_color:   Union[str, int, Tuple[int, int, int], Color],
        sep: str = " ",
        end: str = "\n") -> None:
    """
    print gradients.
    ----------------
    note:
        you must pass both `start_color` and `end_color`.
    raises:
        ValueError: if start_color and end_color are invalid, or unsupported format.

    :param values: any value, will be auto-cast to str.
    :param start_color: start color, this can be in any supported format.
    :param end_color: end color, this can be in any supported format.
    :param sep: separator, gets auto-cast to str.
    :param end: end char, gets auto-cast to str.
    """

    text = str(sep).join(map(str, values))

    start_color = ColorParser.parse(start_color)
    end_color   = ColorParser.parse(end_color)

    steps = len(text)

    grad = gradient(start_color, end_color, steps)

    try:
        for i, char in enumerate(text):
            color = grad[i]
            sys.stdout.write(str(Color(color)) + char)
        sys.stdout.write(str(end))
    finally:
        sys.stdout.write(Fore.RESET)
        sys.stdout.flush()


def gradient_change(
        *values: Any,
        start_color: Union[str, int, Tuple[int, int, int], Color],
        end_color:   Union[str, int, Tuple[int, int, int], Color],
        delay: float = 0.05,
        sep: str = " ",
        end: str = "\n") -> None:
    """
    print gradients and change in place from start_color to end_color in same place. \n
    note:
        both start_color and end_color must be passed.
        THIS DOES NOT WORK WITH MULTILINE TEXTS `YET`. \n
    raises:
        ValueError: if start_color and end_color are invalid, or unsupported format.
    --------------------------------------------------------------------------------
    :param values: any value, will be auto-cast to str.
    :param start_color: start color, this can be in any supported format.
    :param end_color: end color, this can be in any supported format.
    :param delay: delay between each change, recommended to be between 0.005 and 0.1.
    :param sep: separator, gets auto-cast to str.
    :param end: end char, gets auto-cast to str.
    """

    text = str(sep).join(map(str, values))

    start_color = ColorParser.parse(start_color)
    end_color   = ColorParser.parse(end_color)
    steps       = len(text)
    grad        = gradient(start_color, end_color, steps)

    try:
        for idx in range(steps):
            sys.stdout.write(str(Color(grad[idx])) + "\r" + text)
            sys.stdout.flush()
            time.sleep(delay)
        sys.stdout.write(end)
    finally:
        sys.stdout.write(Fore.RESET)
        sys.stdout.flush()


def gradient_scroll(
        *values: Any,
        start_color: Union[str, int, Tuple[int, int, int], Color],
        end_color:   Union[str, int, Tuple[int, int, int], Color],
        delay: float = 0.03,
        times: int = 4,
        reverse: bool = False,
        sep: str = " ",
        end: str = "\n") -> None:
    """
    scroll gradient text horizontally \n
    note:
        both start_color and end_color must be passed. \n
        THIS DOES NOT WORK WITH MULTILINE TEXTS `YET`.  \n
    raises:
        ValueError: if start_color and end_color are invalid, or unsupported format.
    --------------------
    :param values: any value, will be auto-cast to str.
    :param start_color: start color, this can be in any supported format.
    :param end_color: end color, this can be in any supported format.
    :param delay: delay between each change, recommended to be lower than 0.1
    :param times: number of times to scroll
    :param reverse: reverse the gradient
    :param sep: separator, gets auto-cast to str
    :param end: end char, gets auto-cast to str
    """
    text = str(sep).join(map(str, values))
    if len(text) % 2:
        text += " "

    start_color = ColorParser.parse(start_color)
    end_color   = ColorParser.parse(end_color)
    steps       = len(text)

    first_half  = gradient(start_color, end_color, steps // 2)
    second_half = gradient(end_color, start_color, steps // 2)
    grad    = first_half + second_half

    try:
        for _ in range(times * steps):
            grad.append(grad.pop()) if reverse else grad.insert(0, grad.pop())

            for idx, char in enumerate(grad):
                sys.stdout.write(str(Color(char)) + text[idx])
            sys.stdout.write("\r")
            sys.stdout.flush()
            time.sleep(delay)

        sys.stdout.write(end)
    finally:
        sys.stdout.write(Fore.RESET)
        sys.stdout.flush()



import sys
from typing import Any, Union, Tuple
from .parser import *
from .fore import Fore
from .color import Color


__all__ = ["rgbprint"]


def rgbprint(
        *values: Any,
        color: Union[str, int, Tuple[int, int, int], Color, None] = None,
        sep:   str = " ",
        end:   str = "\n") -> None:
    """
    print function with color support.
    -------------------------------------------------------------------------------------------------
    :param values: values to print, this can be Any type. each value is automatically cast to str.
    :param color: color name, color value as string, color hex integer, color tuple or rgbprint.Color object.
    :param sep: separator between values. gets automatically cast to str.
    :param end: end of line. gets automatically cast to str.
    """

    if color is not None:
        color = ColorParser.parse(color)
        color = Color(color)
        sys.stdout.write(str(color))

    text = str(sep).join(map(str, values))

    try:
        sys.stdout.write(text)
    finally:
        sys.stdout.write(str(end))
        sys.stdout.write(Fore.RESET)



from typing import Tuple, Union
from .parser import ColorParser


__all__ = ["Color"]


class Color:
    """
    A class to represent a color.
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    color class.
    can be printed, and will work with the RGB print function.  \n
    """
    def __init__(self, color: Union[str, int, Tuple[int, int, int]]):
        """
        Initialize a color.
        :param color: this can be a string "RRGGBB" or "#RRGGBB",
        a color name (see supported colors), integer color (0xFF00FF) or a (r, g, b) tuple of integers (0-255).

        raises:
            ValueError: if the color is not supported.
        """
        if isinstance(color, Color):
            raise ValueError("Can't create a color from another color! instead use copy()")
        rgb = ColorParser.parse(color)
        self.r, self.g, self.b = rgb

    def __str__(self):
        return "\033[38;2;{0};{1};{2}m".format(self.r, self.g, self.b)

    def __eq__(self, other):
        return self.r == other.r and self.g == other.g and self.b == other.b

    def __repr__(self):
        return "Color({0}, {1}, {2})".format(self.r, self.g, self.b)

    # TODO copy and dunders

    # feature in the future
    #
    # def __add__(self, other):
    #     return Color((self.r + other.r, self.g + other.g, self.b + other.b))
    #
    # def __sub__(self, other):
    #     return Color((self.r - other.r, self.g - other.g, self.b - other.b))
    #
    # def __mul__(self, other):
    #     return Color((self.r * other, self.g * other, self.b * other))
    #
    # def __floordiv__(self, other):
    #     return Color((self.r // other, self.g // other, self.b // other))



import time

from .rgbprint import rgbprint
from .gradient import gradient
from .color import Color
from itertools import cycle
from typing import Generator
from typing import Union, Tuple

__all__ = ["load_bar"]


def load_bar(start_color: Union[str, int, Tuple[int, int, int], Color],
             end_color:   Union[str, int, Tuple[int, int, int], Color],
             *,
             width: int   = 20,
             delay: float = 0.5,
             fill:  str   = " ",
             delimiter: str = "█",
             ) -> Generator:
    """
    generator that prints a load bar with gradients! \n
    raises:
        ValueError when passed an invalid color.
    ------------------------------------------------------
    :param start_color:  the start color of the gradient.
    :param end_color:    the end color of the gradient.
    :param width:        the width of the load bar.
    :param delay:        the delay between each tick.
    :param fill:         the fill character.
    :param delimiter:    the delimiter character.
    """
    start_color = Color(start_color)
    end_color   = Color(end_color)

    delimiter = str(delimiter)
    fill      = str(fill)

    grad      = cycle(gradient((start_color.r, start_color.g, start_color.b),
                               (end_color.r, end_color.g, end_color.b),
                               width))

    for progress in range(width + 1):
        rgbprint("[", color=Color((255, 255, 255)), sep="", end="")

        content = "{}{}".format(delimiter*progress, fill*(width-progress))

        for bar in content:
            rgbprint(bar, color=next(grad), sep="", end="")

        rgbprint("]", color=Color((255, 255, 255)), sep="", end="\r")
        time.sleep(delay)
        yield

    rgbprint()



import string
from typing import Tuple
from .supported_colors import SupportedColors


__all__ = ["ColorParser"]


class ColorParser:
    """
    the color parser class, used to parse: color names, strings, ints, hex ints and tuples into color tuples
    """

    @classmethod
    def parse(cls, color) -> Tuple[int, int, int]:
        """
        parses the string, int, or hex colors into color tuples
        :param color: either color name string, int, hex int, or tuple of the color
        raises:
            ValueError when passed an invalid color
        """
        from .color import Color

        if isinstance(color, str):
            if color.upper() in SupportedColors.__members__:
                return cls._parse_color_name(color)

            if color.startswith("#"):
                color = color.removeprefix("#")
                if all(c in string.hexdigits for c in color):
                    return cls._parse_hex_string(color)

        if isinstance(color, int):
            if color in range(0xFFFFFF+1):
                return cls._parse_int(color)

        if isinstance(color, tuple):
            if all(rgb in range(0xFF+1) for rgb in color):
                return color

        if isinstance(color, Color):
            color = color.r, color.g, color.b

        raise ValueError("{} is not a valid color!".format(color))

    @classmethod
    def _parse_color_name(cls, color: str) -> Tuple[int, int, int]:
        return SupportedColors[color.upper()].value

    @classmethod
    def _parse_hex_string(cls, color: str) -> Tuple[int, int, int]:
        try:
            color = tuple(bytes.fromhex(color))
        except ValueError as exc:
            raise ValueError("{} is not a valid hex color!".format(color)) from exc

        if not all(rgb in range(0xFF+1) for rgb in color) or len(color) != 3:
            raise ValueError("{} is not a valid hex color!".format(color))

        return color

    @classmethod
    def _parse_int(cls, color: int) -> Tuple[int, int, int]:
        blue  =   color %  256
        green = ((color -  blue) // 256) % 256
        red   = ((color -  blue) // 256 ** 2) - green // 256
        return red, green, blue

    @classmethod
    def _color_char(cls, color: Tuple[int, int, int]) -> str:
        return "\033[38;2;{0};{1};{2}m".format(*color)



import setuptools


setuptools.setup(
    name="rgbprint",
    version="3.0.5",
    license="GNU",
    author="ddjerqq",
    author_email="ddjerqq@gmail.com",
    url="https://github.com/ddjerqq/rgbprint",
    keywords="rgb print gradient color colorama",
    description="Print rgb colors and gradients in your terminal. Official github: https://github.com/ddjerqq/rgbprint",
    packages=["rgbprint"]
)



"""
# TODO add docs here
"""


import os

from .fore     import *
from .rgbprint import *
from .gradient import *
from .parser   import ColorParser
from .color    import *
from .load_bar import *

__all__ = [
    "Fore",
    "Color",
    "ColorParser",
    "rgbprint",
    "gradient_print",
    "gradient_change",
    "gradient_scroll",
    "load_bar",
]

__version__ = "3.0.5"

os.system("")
# initialize colors lol
